
from abc import abstractmethod
from enum import IntEnum

from sqlalchemy import ForeignKey, Column, Integer, func, and_, UniqueConstraint
from sqlalchemy.event import listens_for
from sqlalchemy.orm import Session, aliased

from ..support import Base
from ..types import OpenSched, Date, ShortCls, short_cls
from ...lib.date import to_time, time_to_local_date, max_time, min_time, extend_range


class SourceType(IntEnum):

    SOURCE = 0
    INTERVAL = 1
    ACTIVITY = 2
    TOPIC = 3
    CONSTANT = 4
    MONITOR = 5
    SEGMENT = 6


class Source(Base):

    __tablename__ = 'source'

    id = Column(Integer, primary_key=True)
    type = Column(Integer, nullable=False, index=True)  # index needed for fast deletes of subtypes

    __mapper_args__ = {
        'polymorphic_identity': SourceType.SOURCE,
        'polymorphic_on': type
    }

    @abstractmethod
    def time_range(self, s):
        raise NotImplementedError('time_range for %s' % self)

    @classmethod
    def before_flush(cls, s):
        cls.__clean_dirty_intervals(s)

    @classmethod
    def __clean_dirty_intervals(cls, s):
        from .statistic import StatisticJournal
        # sessions are generally restricted to one time region, so we'll bracket that rather
        # than list all times
        start, finish = None, None
        # all sources except intervals that are being deleted (need to catch on cascade to statistics)
        for instance in s.deleted:
            if isinstance(instance, Source) and not isinstance(instance, Interval):
                a, b = instance.time_range(s)
                start, finish = min_time(a, start), max_time(b, finish)
        # all modified statistics
        for instance in s.dirty:
            # ignore constants as time 0
            if isinstance(instance, StatisticJournal) and s.is_modified(instance) and instance.time:
                start, finish = extend_range(start, finish, instance.time)
        # all new statistics that aren't associated with intervals and have non-null data
        # (avoid triggering on empty diary entries)
        for instance in s.new:
            # ignore constants as time 0
            if isinstance(instance, StatisticJournal) and not isinstance(instance.source, Interval) \
                    and instance.value is not None and instance.time:
                start, finish = extend_range(start, finish, instance.time)
        if start is not None:
            Interval.clean_times(s, start, finish)


@listens_for(Session, 'before_flush')
def before_flush(session, context, instances):
    Source.before_flush(session)


class NoStatistics(Exception):
    pass


class Interval(Source):

    __tablename__ = 'interval'

    id = Column(Integer, ForeignKey('source.id', ondelete='cascade'), primary_key=True)
    schedule = Column(OpenSched, nullable=False, index=True)
    # disambiguate creator so each can wipe only its own data on force
    owner = Column(ShortCls, nullable=False)
    # these are for the schedule - finish is redundant (start is not because of timezone issues)
    start = Column(Date, nullable=False, index=True)
    finish = Column(Date, nullable=False, index=True)
    UniqueConstraint(schedule, owner, start)

    __mapper_args__ = {
        'polymorphic_identity': SourceType.INTERVAL
    }

    def __str__(self):
        owner = self.owner if isinstance(self.owner, str) else short_cls(self.owner)
        return 'Interval "%s from %s" (owner %s)' % (self.schedule, self.start, owner)

    @classmethod
    def _missing_interval_start_dates(cls, log, s, schedule, interval_owner, statistic_owner=None):
        '''
        Returns a list of times that mark the start DATES of missing intervals covering "all"
        statistics (optionally with a given owner and excluding constants at "zero time"),
        plus the end TIME of the statistics,
        '''
        stats_start, stats_finish = cls._raw_statistics_time_range(s, statistic_owner)
        log.debug('Statistics exist %s - %s' % (stats_start, stats_finish))
        starts = cls._open_interval_dates(s, schedule, interval_owner)
        stats_start_date = schedule.start_of_frame(time_to_local_date(stats_start))
        if not starts or starts[0] > stats_start_date:
            if not cls._existing_interval_at(s, schedule, interval_owner, stats_start_date):
                starts = [stats_start_date] + starts
        log.debug('Have %d open blocks finishing at %s' % (len(starts), stats_finish))
        for i, start in enumerate(starts):
            log.debug('Block %d starts at %s' % (i, start))
        log.debug('')
        return starts, stats_finish

    @classmethod
    def _raw_statistics_time_range(cls, s, statistics_owner=None):
        '''
        The time range over which statistics exist (optionally restricted by owner),
        ignoring constants are "time zero".  This is the first to the last time for
        any statistics - it pays no attention to gaps.
        '''
        from .statistic import StatisticJournal, StatisticName
        q = s.query(func.min(StatisticJournal.time), func.max(StatisticJournal.time)). \
            filter(StatisticJournal.time > to_time(24 * 60 * 60.0))
        if statistics_owner:
            q = q.join(StatisticName).filter(StatisticName.owner == statistics_owner)
        start, finish = q.one()   # skip entire first day because tz
        if start and finish:
            return start, finish
        else:
            raise NoStatistics('No statistics are currently defined')

    @classmethod
    def _open_interval_dates(cls, s, schedule, interval_owner):
        '''
        The finish DATES that come at the ends of contiguous intervals
        (for intervals of the given schedule and owner)
        '''
        close = aliased(Interval)
        return [result[0] for result in s.query(Interval.finish). \
            outerjoin(close,
                      and_(Interval.finish == close.start,
                           Interval.owner == close.owner,
                           Interval.schedule == close.schedule)). \
            filter(close.start == None,
                   Interval.owner == interval_owner,
                   Interval.schedule == schedule). \
            order_by(Interval.finish).all()]

    @classmethod
    def _existing_interval_at(cls, s, schedule, interval_owner, start):
        '''
        The existing interval for a given start, owner, schedule.
        '''
        return s.query(Interval). \
            filter(Interval.start == start,
                   Interval.schedule == schedule,
                   Interval.owner == interval_owner).one_or_none()

    @classmethod
    def _missing_interval_dates_from(cls, log, s, schedule, interval_owner, start, finish):
        '''
        Iterator over start DATES for intervals that are not present in the database,
        starting from `start` and ending when an interval is found, or with `finish` if no intervals are found.
        '''
        while start <= time_to_local_date(finish):
            next = schedule.next_frame(start)
            log.debug('Missing Interval %s - %s' % (start, next))
            yield start, next
            start = next
            if cls._existing_interval_at(s, schedule, interval_owner, start):
                return

    @classmethod
    def first_missing_date(cls, log, s, schedule, interval_owner, statistic_owner=None):
        '''
        For Impulse (and anything else that needs to delete forwards)
        '''
        starts, overall_finish = cls._missing_interval_start_dates(log, s, schedule, interval_owner, statistic_owner)
        if starts:
            return starts[0], time_to_local_date(overall_finish)
        else:
            return None, None

    @classmethod
    def missing_dates(cls, log, s, schedule, interval_owner, statistic_owner=None):
        '''
        Iterator over start DATES for all missing intervals, given the constraints supplied.
        '''
        starts, overall_finish = cls._missing_interval_start_dates(log, s, schedule, interval_owner, statistic_owner)
        for block_start in starts:
            yield from cls._missing_interval_dates_from(log, s, schedule, interval_owner, block_start, overall_finish)

    @classmethod
    def delete_all(cls, log, s):
        '''
        Efficiently delete all intervals.
        '''
        log.warning('Deleting all Intervals')
        # this uses the on delete cascade between source and interval
        s.query(Source).filter(Source.type == SourceType.INTERVAL).delete()

    @classmethod
    def clean_times(cls, s, start, finish, owner=None):
        '''
        Remove all intervals that include data in the given TIME range,
        '''
        cls.clean_dates(s, time_to_local_date(start), time_to_local_date(finish), owner=owner)

    @classmethod
    def clean_dates(cls, s, start, finish, owner=None):
        '''
        Remove all summary intervals (not monitor intervals) in the given DATE range.
        '''
        q = s.query(Interval).filter(Interval.start < finish, Interval.finish >= start)
        if owner:
            q = q.filter(Interval.owner == owner)
        for interval in q.all():
            s.delete(interval)




from random import choice
from re import split

from sqlalchemy import func, inspect, and_, select
from sqlalchemy.sql.functions import coalesce

from . import IntervalCalculator
from ..names import MAX, MIN, SUM, CNT, AVG, MSR, ENTRIES
from ...lib.date import local_date_to_time
from ...squeal.database import add
from ...squeal.tables.source import Interval
from ...squeal.tables.statistic import StatisticJournal, StatisticName, StatisticMeasure, StatisticJournalInteger, \
    StatisticJournalFloat, StatisticJournalText, TYPE_TO_JOURNAL_CLASS


class SummaryStatistics(IntervalCalculator):

    def _filter_intervals(self, q, schedule=None):
        return q.filter(Interval.schedule == schedule,
                        Interval.owner == self)

    def _statistics_missing_summaries(self, s, start, finish):
        statistics_with_data_but_no_summary = s.query(StatisticName.id). \
            join(StatisticJournal). \
            filter(StatisticJournal.time >= start,
                   StatisticJournal.time < finish,
                   StatisticName.summary != None)
        # avoid duplicates
        return s.query(StatisticName). \
            filter(StatisticName.id.in_(statistics_with_data_but_no_summary)). \
            all()

    def _calculate_value(self, s, statistic_name, summary, pessimistic, start_time, finish_time, interval, schedule):

        sj = inspect(StatisticJournal).local_table
        sji = inspect(StatisticJournalInteger).local_table
        sjf = inspect(StatisticJournalFloat).local_table
        sjt = inspect(StatisticJournalText).local_table

        units = statistic_name.units
        values = coalesce(sjf.c.value, sji.c.value, sjt.c.value)

        if summary == MAX:
            result = func.max(values)
        elif summary == MIN:
            result = func.min(values)
        elif summary == SUM:
            result = func.sum(values)
        elif summary == CNT:
            result = func.count(values)
            units = ENTRIES
        elif summary == AVG:
            result = func.avg(values)
        elif summary == MSR:
            self._calculate_measures(s, statistic_name, pessimistic, start_time, finish_time, interval)
            return False
        else:
            raise Exception('Bad summary: %s' % summary)

        stmt = select([result]). \
            select_from(sj.outerjoin(sjf).outerjoin(sji).outerjoin(sjt)). \
            where(and_(sj.c.statistic_name_id == statistic_name.id,
                       sj.c.time >= start_time,
                       sj.c.time < finish_time))

        value = next(s.connection().execute(stmt))[0]

        if value is not None:
            name = self.fmt_name(statistic_name.name, summary, schedule)
            jtype = TYPE_TO_JOURNAL_CLASS[type(value)]
            new_name = StatisticName.add_if_missing(self._log, s, name, units, None, self, statistic_name)
            journal = add(s, jtype(statistic_name=new_name, source=interval, value=value, time=start_time))
            self._log.debug('Created %s over %s for %s' % (journal, interval, statistic_name))

        return bool(value)

    def _calculate_measures(self, s, statistic_name, pessimistic, start_time, finish_time, interval):

        data = sorted([x for x in
                       s.query(StatisticJournal).
                      filter(StatisticJournal.statistic_name == statistic_name,
                             StatisticJournal.time >= start_time,
                             StatisticJournal.time < finish_time).all()
                       if x is not None], key=lambda x: x.value, reverse=not pessimistic)

        n, measures = len(data), []
        for rank, journal in enumerate(data, start=1):
            if n > 1:
                percentile = (n - rank) / (n - 1) * 100
            else:
                percentile = 100
            measure = StatisticMeasure(statistic_journal=journal, source=interval, rank=rank, percentile=percentile)
            s.add(measure)
            measures.append(measure)
        if n > 8:  # avoid overlap in fuzzing (and also, plot individual points in this case)
            for q in range(5):
                measures[fuzz(n, q)].quartile = q
        self._log.debug('Ranked %s' % statistic_name)

    def _run_calculations(self, schedule):
        with self._db.session_context() as s:
            for start, finish in Interval.missing_dates(self._log, s, schedule, self):
                start_time, finish_time = local_date_to_time(start), local_date_to_time(finish)
                interval = add(s, Interval(start=start, finish=finish, schedule=schedule, owner=self))
                have_data = False
                for statistic_name in self._statistics_missing_summaries(s, start_time, finish_time):
                    self._log.debug(statistic_name)
                    summaries = [x.lower() for x in split(r'[\s,]*(\[[^\]]+\])[\s ]*', statistic_name.summary) if x]
                    pessimistic = MIN in summaries
                    for summary in summaries:
                        have_data |= self._calculate_value(s, statistic_name, summary, pessimistic,
                                                           start_time, finish_time, interval, schedule)
                if have_data:
                    self._log.info('Added statistics for %s' % interval)
                else:
                    s.delete(interval)

    @classmethod
    def parse_name(cls, name):
        left, right = name.split(' ', 1)
        summary, period = left.split('/')
        return summary, period, right

    @classmethod
    def fmt_name(cls, name, summary, schedule):
        title = summary[1:-1].capitalize()   # see parse_name
        return '%s/%s %s' % (title, schedule.describe(), name)


def fuzz(n, q):
    i = (n-1) * q / 4
    if i != int(i):
        i = int(i) + choice([0, 1])  # if we're between two points, pick either
    return int(i)




from sqlalchemy import Column, Integer, ForeignKey, Float, UniqueConstraint
from sqlalchemy.orm import relationship, backref

from ..support import Base
from ..types import Str


class ActivitySimilarity(Base):

    __tablename__ = 'activity_similarity'

    id = Column(Integer, primary_key=True)
    constraint = Column(Str, index=True)
    activity_journal_lo_id = Column(Integer, ForeignKey('activity_journal.id', ondelete='cascade'), index=True)
    activity_journal_lo = relationship('ActivityJournal', foreign_keys=[activity_journal_lo_id])
    activity_journal_hi_id = Column(Integer, ForeignKey('activity_journal.id', ondelete='cascade'), index=True)
    activity_journal_hi = relationship('ActivityJournal', foreign_keys=[activity_journal_hi_id])
    similarity = Column(Float, nullable=False)
    UniqueConstraint(constraint, activity_journal_lo_id)
    UniqueConstraint(constraint, activity_journal_hi_id)


class ActivityNearby(Base):

    __tablename__ = 'activity_nearby'

    id = Column(Integer, primary_key=True)
    constraint = Column(Str, index=True)
    group = Column(Integer, nullable=False, index=True)
    activity_journal_id = Column(Integer, ForeignKey('activity_journal.id', ondelete='cascade'))
    activity_journal = relationship('ActivityJournal',
                                    backref=backref('nearby', cascade='all, delete-orphan',
                                                    passive_deletes=True))
    UniqueConstraint(constraint, activity_journal_id)




from subprocess import run
from tempfile import NamedTemporaryFile

from sqlalchemy.sql.functions import count

from ch2.command.args import m, V, bootstrap_file
from ch2.config.personal import acooke
from ch2.lib.date import to_time, local_date_to_time, to_date
from ch2.squeal.database import add
from ch2.squeal.tables.source import Source, Interval
from ch2.squeal.tables.statistic import StatisticJournalText, StatisticJournal, StatisticJournalFloat, StatisticName, \
    StatisticJournalInteger, StatisticJournalType
from ch2.squeal.tables.topic import TopicJournal, Topic
from ch2.stoats.calculate.summary import SummaryStatistics


# the idea here is to test the new database schema with sources etc
# so we configure a database then load some data, calculate some stats,
# and see if everything works as expected.


def test_sources():

    with NamedTemporaryFile() as f:

        args, log, db = bootstrap_file(f, m(V), '5', configurator=acooke)

        with db.session_context() as s:

            # add a diary entry

            diary = s.query(Topic).filter(Topic.name == 'Diary').one()
            d = add(s, TopicJournal(topic=diary, date='2018-09-29'))
            d.populate(log, s)
            assert len(d.topic.fields) == 9, list(enumerate(map(str, d.topic.fields)))
            assert d.topic.fields[0].statistic_name.name == 'Notes'
            assert d.topic.fields[1].statistic_name.name == 'Weight', str(d.topic.fields[1])
            for field in d.topic.fields:
                if field in d.statistics:
                    assert d.statistics[field].value is None, field
            d.statistics[d.topic.fields[0]].value = 'hello world'
            d.statistics[d.topic.fields[1]].value = 64.5

        with db.session_context() as s:

            # check the diary entry was persisted

            diary = s.query(Topic).filter(Topic.name == 'Diary').one()
            d = s.query(TopicJournal).filter(TopicJournal.topic == diary,
                                             TopicJournal.date == '2018-09-29').one()
            s.flush()
            d.populate(log, s)
            assert len(d.topic.fields) == 9, list(enumerate(map(str, d.topic.fields)))
            assert d.topic.fields[0].statistic_name.name == 'Notes'
            assert d.statistics[d.topic.fields[0]].value == 'hello world'
            assert d.topic.fields[1].statistic_name.name == 'Weight'
            assert d.statistics[d.topic.fields[1]].value == 64.5
            assert d.statistics[d.topic.fields[1]].type == StatisticJournalType.FLOAT

        # generate summary stats

        SummaryStatistics(log, db, schedule='m').run()
        SummaryStatistics(log, db, schedule='y').run()

        with db.session_context() as s:

            # check the summary stats

            diary = s.query(Topic).filter(Topic.name == 'Diary').one()
            weight = s.query(StatisticJournal).join(StatisticName). \
                filter(StatisticName.owner == diary, StatisticName.name == 'Weight').one()
            assert weight.value == 64.5
            assert len(weight.measures) == 2, weight.measures
            assert weight.measures[0].rank == 1
            assert weight.measures[0].percentile == 100, weight.measures[0].percentile
            n = s.query(count(StatisticJournalFloat.id)).scalar()
            assert n == 4, n
            n = s.query(count(StatisticJournalInteger.id)).scalar()
            assert n == 11, n
            m_avg = s.query(StatisticJournalFloat).join(StatisticName). \
                filter(StatisticName.name == 'Avg/Month Weight').one()
            assert m_avg.value == 64.5
            y_avg = s.query(StatisticJournalFloat).join(StatisticName). \
                filter(StatisticName.name == 'Avg/Year Weight').one()
            assert y_avg.value == 64.5
            month = s.query(Interval).filter(Interval.schedule == 'm').one()
            assert month.start == to_date('2018-09-01'), month.start
            assert month.finish == to_date('2018-10-01'), month.finish

        with db.session_context() as s:

            # delete the diary entry

            diary = s.query(Topic).filter(Topic.name == 'Diary').one()
            d = s.query(TopicJournal).filter(TopicJournal.topic == diary,
                                             TopicJournal.date == '2018-09-29').one()
            s.delete(d)

        run('sqlite3 %s ".dump"' % f.name, shell=True)

        with db.session_context() as s:

            # check the delete cascade

            assert s.query(count(TopicJournal.id)).scalar() == 0
            # this should be zero because the Intervals were automatically deleted
            for source in s.query(Source).all():
                print(source)
            for journal in s.query(StatisticJournal).all():
                print(journal)
            assert s.query(count(Source.id)).scalar() == 6, list(map(str, s.query(Source).all()))  # constants
            # 3 JSON entries for impulse + 1 for nearby
            assert s.query(count(StatisticJournalText.id)).scalar() == 4, s.query(count(StatisticJournalText.id)).scalar()
            assert s.query(count(StatisticJournal.id)).scalar() == 4, s.query(count(StatisticJournal.id)).scalar()




from logging import getLogger, DEBUG, Formatter, INFO, StreamHandler, NullHandler
from logging.handlers import RotatingFileHandler
from os.path import join

from ..command.args import COMMAND, LOGS, PROGNAME, VERBOSITY


CACHE = []


def make_log(args, tui=False):

    if not CACHE:

        level_unset = args[VERBOSITY] is None
        level = 4 if level_unset else args[VERBOSITY][0]
        level = 10 * (6 - level)

        file_formatter = Formatter('%(levelname)-8s %(asctime)s: %(message)s')
        name = args[COMMAND] if COMMAND in args else PROGNAME
        path = join(args.dir(LOGS), name + '.log')
        file_handler = RotatingFileHandler(path, maxBytes=1e6, backupCount=10)
        file_handler.setLevel(DEBUG)
        file_handler.setFormatter(file_formatter)

        slog = getLogger('sqlalchemy')
        slog.setLevel(INFO)
        slog.addHandler(file_handler)

        mlog = getLogger('matplotlib')
        mlog.setLevel(INFO)
        mlog.addHandler(file_handler)

        blog = getLogger('bokeh')
        blog.setLevel(INFO)
        blog.addHandler(file_handler)

        tlog = getLogger('tornado')
        tlog.setLevel(INFO)
        tlog.addHandler(file_handler)

        log = getLogger(name)
        log.setLevel(DEBUG)
        log.addHandler(file_handler)

        if not tui or not level_unset:
            stderr_formatter = Formatter('%(levelname)8s: %(message)s')
            stderr_handler = StreamHandler()
            stderr_handler.setLevel(level)
            stderr_handler.setFormatter(stderr_formatter)
            log.addHandler(stderr_handler)
            # slog.addHandler(stderr_handler)
            # mlog.addHandler(stderr_handler)
            # blog.addHandler(stderr_handler)
            # tlog.addHandler(stderr_handler)

        CACHE.append(log)

    return CACHE[0]




import pandas as pd
from bokeh.models import Range1d


def interpolate_to(df, name):
    df2 = df.copy()
    df2['keep'] = pd.notna(df2[name])
    df2.interpolate(method='time', inplace=True)
    df2 = df2.loc[df2['keep'] == True]
    return df2


def interpolate_to_index(reference, raw, method='linear'):
    reference = pd.DataFrame(True, index=reference.index, columns=['keep'])
    raw = pd.DataFrame(raw, index=raw.index)
    both = reference.merge(raw, how='outer', left_index=True, right_index=True)
    both.loc[pd.isna(both['keep']), ['keep']] = False
    both.interpolate(method=method, inplace=True)
    both = both.loc[both['keep'] == True]
    return both.drop(columns=['keep']).dropna().iloc[:, 0]


def closed_patch(y, zero=0):
    y = y.dropna()
    x = y.index
    return y.append(pd.Series([zero, zero], index=[x[len(x)-1], x[0]]))


def _delta(z):
    if len(z.dropna()):
        return closed_patch(z)
    else:
        return None


def delta_patches(y1, y2):
    dy = y1 - y2
    scale = dy.dropna().abs().max() * 1.1
    y1 = _delta(dy.clip(lower=0))
    y2 = _delta(dy.clip(upper=0))
    return y1, y2, Range1d(start=-scale, end=scale)




from tempfile import NamedTemporaryFile

import sqlalchemy.sql.functions as func

from ch2 import monitor
from ch2.command.args import bootstrap_file, m, V, DEV, mm, FAST
from ch2.config.default import default
from ch2.lib.date import to_time, to_date, local_date_to_time
from ch2.squeal.tables.monitor import MonitorJournal
from ch2.squeal.tables.pipeline import PipelineType
from ch2.squeal.tables.source import Source, SourceType, Interval
from ch2.squeal.tables.statistic import StatisticJournal, StatisticName
from ch2.stoats.calculate import run_pipeline_after
from ch2.stoats.calculate.monitor import MonitorStatistics
from ch2.stoats.names import STEPS, REST_HR, DAILY_STEPS


def test_monitor():

    with NamedTemporaryFile() as f:

        args, log, db = bootstrap_file(f, m(V), '5')

        bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)

        args, log, db = bootstrap_file(f, m(V), '5', mm(DEV),
                                       'monitor', mm(FAST), 'data/test/personal/25822184777.fit')
        monitor(args, log, db)

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        run_pipeline_after(log, db, PipelineType.STATISTIC, force=True, after='2018-01-01')

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        with db.session_context() as s:
            n = s.query(func.count(StatisticJournal.id)).scalar()
            assert n == 109, n
            mjournal = s.query(MonitorJournal).one()
            assert mjournal.start != mjournal.finish


def test_values():

    with NamedTemporaryFile() as f:

        args, log, db = bootstrap_file(f, m(V), '5')

        bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)

        for file in ('24696157869', '24696160481', '24696163486'):
            args, log, db = bootstrap_file(f, m(V), '5', mm(DEV),
                                           'monitor', mm(FAST), 'data/test/personal/andrew@acooke.org_%s.fit' % file)
            monitor(args, log, db)

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        run_pipeline_after(log, db, PipelineType.STATISTIC, force=True, after='2018-01-01')

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        with db.session_context() as s:

            mjournals = s.query(MonitorJournal).order_by(MonitorJournal.start).all()
            assert mjournals[2].start == to_time('2018-09-06 15:06:00'), mjournals[2].start
            print(mjournals[2].fit_file)

            # steps

            summary = s.query(StatisticJournal).join(StatisticName, Interval). \
                filter(Interval.start == '2018-09-06',
                       Interval.schedule == 'd',
                       StatisticName.owner == MonitorStatistics,
                       StatisticName.name == DAILY_STEPS).one()
            # connect has 12757 for this date,
            assert summary.value == 12757, summary.value

            # heart rate

            summary = s.query(StatisticJournal).join(StatisticName, Interval). \
                filter(Interval.start == '2018-09-06',
                       Interval.schedule == 'd',
                       StatisticName.owner == MonitorStatistics,
                       StatisticName.name == REST_HR).one()
            assert summary.value == 42, summary.value


FILES = ('25505915679', '25519562859', '25519565531', '25532154264', '25539076032', '25542112328')


def generic_bug(files):

    with NamedTemporaryFile() as f:

        args, log, db = bootstrap_file(f, m(V), '5')

        bootstrap_file(f, m(V), '5', mm(DEV), configurator=default)

        for file in files:
            args, log, db = bootstrap_file(f, m(V), '5', mm(DEV),
                                           'monitor', mm(FAST), 'data/test/personal/andrew@acooke.org_%s.fit' % file)
            monitor(args, log, db)

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        run_pipeline_after(log, db, PipelineType.STATISTIC, force=True, after='2018-01-01')

        # run('sqlite3 %s ".dump"' % f.name, shell=True)

        with db.session_context() as s:

            # steps

            summary = s.query(StatisticJournal).join(StatisticName, Interval). \
                filter(Interval.start == '2018-10-07',
                       Interval.schedule == 'd',
                       StatisticName.owner == MonitorStatistics,
                       StatisticName.name == DAILY_STEPS).one()

            # connect has 3031 for this date.
            assert summary.value == 3031, summary.value


def test_bug():
    generic_bug(sorted(FILES))


def test_bug_reversed():
    generic_bug(sorted(FILES, reverse=True))
